import numpy as np
from gammapy.datasets import MapDataset

from titrate.utils import copy_models_to_dataset


class AsimovMapDataset(MapDataset):
    """
    AsimovMapDataset is a subclass of a gammapy MapDataset
    and provides asimov-like fake method.
    """

    def fake(self):
        """
        This method generates Asimov like counts,
        i.e. the counts are not drawn from a poisson distribution.
        """
        npred = self.npred()
        data = np.nan_to_num(npred.data, copy=True, nan=0.0, posinf=0.0, neginf=0.0)
        npred.data = data
        self.counts = npred

    @classmethod
    def from_MapDataset(self, dataset):
        dataset_dict = dataset.__dict__.copy()

        delete_keys = [key for key in dataset_dict.keys() if key.startswith("_")]

        deleted_entries = {}
        for key in delete_keys:
            deleted_entries[key] = dataset_dict.pop(key)

        asimov_dataset = AsimovMapDataset(**dataset_dict)
        # copy IRFs separately
        asimov_dataset._background_parameters_cached = deleted_entries[
            "_background_parameters_cached"
        ]

        copy_models_to_dataset(dataset.models, asimov_dataset)
        asimov_dataset.fake()

        return asimov_dataset



import numpy as np
from gammapy.modeling import Fit
from scipy.interpolate import interp1d
from scipy.optimize import brentq
from scipy.stats import norm

from titrate.datasets import AsimovMapDataset
from titrate.statistics import QMuTestStatistic, QTildeMuTestStatistic

STATISTICS = {"qmu": QMuTestStatistic, "qtildemu": QTildeMuTestStatistic}


class ULCalculator:
    def __init__(
        self, measurement_dataset, statistic="qmu", poi_name="", cl=0.95, cl_type="s"
    ):
        self.measurement_dataset = measurement_dataset
        self.poi_name = poi_name
        if statistic not in STATISTICS.keys():
            raise ValueError(
                "Statistic must be one of {}".format(list(STATISTICS.keys()))
            )
        self.statistic = STATISTICS[statistic](
            self.measurement_dataset, poi_name=self.poi_name
        )

        if cl_type not in ["s+b", "s"]:
            raise ValueError('cl_type must be either "s+b" or "s"')
        self.cl_type = cl_type
        self.cl = cl

    def compute(self):
        poi_best = self.statistic.poi_best

        if poi_best < 0:
            poi_ul = 1e-2
        else:
            poi_ul = poi_best + 0.01 * poi_best
        while self.pvalue(poi_ul, cl_type=self.cl_type) > 1 - self.cl:
            poi_ul *= 2

        interp_ul_points = np.linspace(poi_ul / 10, poi_ul, 10)
        interp_pvalues = np.array(
            [
                self.pvalue(interp_ul, cl_type=self.cl_type)
                for interp_ul in interp_ul_points
            ]
        ).ravel()
        interpolation = interp1d(interp_ul_points, interp_pvalues - 1 + self.cl)
        poi_ul = brentq(interpolation, poi_ul / 10, poi_ul)

        return poi_ul

    def pvalue(self, poi_ul, cl_type):
        pval_bkg = 0
        if self.statistic.__class__.__name__ == "QTildeMuTestStatistic":
            asimov_dataset = AsimovMapDataset.from_MapDataset(self.measurement_dataset)
            asimov_dataset.models.parameters[self.poi_name].value = poi_ul
            asimov_dataset.fake()
            statistic = self.statistic.__class__(asimov_dataset, poi_name=self.poi_name)
            ts_val = self.statistic.evaluate(poi_ul)  # ts_val on measurement_dataset

            pval_sig_bkg = statistic.pvalue(poi_ul, ts_val=ts_val)

            if cl_type == "s":
                no_signal_asimov_dataset = AsimovMapDataset.from_MapDataset(
                    self.measurement_dataset
                )
                no_signal_asimov_dataset.models.parameters[self.poi_name].value = 0
                no_signal_asimov_dataset.fake()
                no_signal_statistic = self.statistic.__class__(
                    no_signal_asimov_dataset, poi_name=self.poi_name
                )
                ts_val_no_signal = self.statistic.evaluate(
                    0
                )  # ts_val on measurement_dataset with no signal

                pval_bkg = no_signal_statistic.pvalue(0, ts_val=ts_val_no_signal)

        else:
            pval_sig_bkg = self.statistic.pvalue(poi_ul)

            if cl_type == "s":
                pval_bkg = self.statistic.pvalue(0)

        return pval_sig_bkg / (1 - pval_bkg)

    def expected_uls(self):
        # Create asimov dataset
        asimov_dataset = AsimovMapDataset.from_MapDataset(self.measurement_dataset)
        asimov_dataset.models.parameters[self.poi_name].value = 0
        asimov_dataset.fake()

        fit = Fit()
        fit_result = fit.run(datasets=[asimov_dataset])

        sigma = np.sqrt(fit_result.covariance_result.matrix[0, 0])
        med_ul = self.compute_band(sigma, 0, self.cl_type)

        one_sig_plus = self.compute_band(sigma, 1, self.cl_type)
        one_sig_minus = self.compute_band(sigma, -1, self.cl_type)

        two_sig_plus = self.compute_band(sigma, 2, self.cl_type)
        two_sig_minus = self.compute_band(sigma, -2, self.cl_type)

        return {
            "med": med_ul,
            "1sig": [one_sig_minus, one_sig_plus],
            "2sig": [two_sig_minus, two_sig_plus],
        }

    def compute_band(self, sigma, n_sigma, cl_type):
        if cl_type == "s+b":
            return sigma * (norm.ppf(self.cl) + n_sigma)
        elif cl_type == "s":
            return sigma * (norm.ppf(1 - (1 - self.cl) * norm.pdf(n_sigma)) + n_sigma)



from functools import lru_cache

import matplotlib.pyplot as plt
import numpy as np
from joblib import Parallel, delayed

from titrate.statistics import QMuTestStatistic, QTildeMuTestStatistic, kstest
from titrate.utils import calc_ts_toyMC

STATISTICS = {"qmu": QMuTestStatistic, "qtildemu": QTildeMuTestStatistic}


class AsymptoticValidator:
    def __init__(
        self, measurement_dataset, asimov_dataset, statistic="qmu", poi_name=""
    ):
        if statistic not in STATISTICS.keys():
            raise ValueError(
                "Statistic must be one of {}".format(list(STATISTICS.keys()))
            )
        self.statistic_key = statistic
        self.statistic = STATISTICS[statistic]
        self.measurement_dataset = measurement_dataset
        self.asimov_dataset = asimov_dataset
        self.poi_name = poi_name

    def validate(self, n_toys=1000):
        toys_ts_diff = self.toys_ts(n_toys, 1, 0)
        toys_ts_same = self.toys_ts(n_toys, 1, 1)

        # only validate ts values above zero because
        # QTildeMuTestStatistic cdf will have problems with negative values in sqrt
        toys_ts_diff = toys_ts_diff[toys_ts_diff >= 0]
        toys_ts_same = toys_ts_same[toys_ts_same >= 0]

        stat = self.statistic(self.asimov_dataset, self.poi_name)
        ks_diff = kstest(
            toys_ts_diff,
            lambda x: stat.asympotic_approximation_cdf(
                poi_val=1, same=False, poi_true_val=0, ts_val=x
            ),
        )
        ks_same = kstest(
            toys_ts_same,
            lambda x: stat.asympotic_approximation_cdf(poi_val=1, ts_val=x),
        )

        valid = ks_diff > 0.05 and ks_same > 0.05

        return {"pvalue_diff": ks_diff, "pvalue_same": ks_same, "valid": valid}

    @lru_cache
    def toys_ts(self, n_toys, poi_val, poi_true_val):
        toys_ts = Parallel(n_jobs=-1, verbose=0)(
            delayed(calc_ts_toyMC)(
                self.measurement_dataset,
                self.statistic,
                poi_val,
                poi_true_val,
                self.poi_name,
            )
            for _ in range(n_toys)
        )

        # to ndarray
        toys_ts = np.array(toys_ts).ravel()

        return toys_ts

    def plot_validation(self, n_toys=1000):
        toys_ts_diff = self.toys_ts(n_toys, 1, 0)
        toys_ts_same = self.toys_ts(n_toys, 1, 1)

        max_q = max(toys_ts_diff.max(), toys_ts_same.max())
        bins = np.linspace(0, max_q, 31)
        plt.hist(
            toys_ts_diff,
            bins=bins,
            density=True,
            histtype="step",
            color="tab:blue",
            label=r"$f(q_\mu\vert\mu^\prime)$, poi_val=1, poi_true_val=0",
        )
        plt.hist(
            toys_ts_same,
            bins=bins,
            density=True,
            histtype="step",
            color="tab:orange",
            label=r"$f(q_\mu\vert\mu)$, poi_val=1, poi_true_val=1",
        )

        lin_q = np.linspace(0, max_q, 1000)
        stat = self.statistic(self.asimov_dataset, self.poi_name)

        plt.plot(
            lin_q,
            stat.asympotic_approximation_pdf(
                poi_val=1, same=False, poi_true_val=0, ts_val=lin_q
            ),
            color="tab:blue",
            label=r"$f(q_\mu\vert\mu^\prime)$, asympotic",
        )
        plt.plot(
            lin_q,
            stat.asympotic_approximation_pdf(poi_val=1, ts_val=lin_q),
            color="tab:orange",
            label=r"$f(q_\mu\vert\mu)$, asympotic",
        )

        plt.yscale("log")
        plt.xlim(0, max_q)

        plt.ylabel("pdf")
        plt.xlabel("q")
        plt.title(self.statistic.__name__)
        plt.legend()
        plt.show()
